"""
==========================================================================
OutputUnitRTL.py
==========================================================================
RTL implementation of OutputUnit.

Author : Yanghui Ou, Cheng Tan
  Date : Feb 28, 2019
"""
from pymtl3 import *
from pymtl3.stdlib.ifcs import GetIfcRTL, SendIfcRTL
from pymtl3.stdlib.rtl.queues import NormalQueueRTL


class OutputUnitRTL( Component ):
  def construct( s, PacketType, QueueType=None ):

    # Interface
    s.get  = GetIfcRTL ( PacketType )
    s.send = SendIfcRTL( PacketType )

    s.QueueType = QueueType

    # If no queue type is assigned
    if s.QueueType != None:

      # Component
      s.queue = QueueType( PacketType )

      # Connections
      s.get.ret       //= s.queue.enq.msg
      s.queue.deq.ret //= s.send.msg

      @s.update
      def up_get_deq():
        s.get.en       = s.get.rdy & s.queue.enq.rdy
        s.queue.enq.en = s.get.rdy & s.queue.enq.rdy

      @s.update
      def up_deq_send():
        s.send.en      = s.send.rdy & s.queue.deq.rdy
        s.queue.deq.en = s.send.rdy & s.queue.deq.rdy

    # No ouput queue
    else:

      s.get.ret //= s.send.msg

      @s.update
      def up_get_send():
        s.get.en  = s.get.rdy & s.send.rdy
        s.send.en = s.get.rdy & s.send.rdy

  def line_trace( s ):
    if s.QueueType != None:
      return "{}({}){}".format( s.get, s.queue.count, s.send )
    else:
      return "{}(0){}".format( s.get, s.send)



'''
==========================================================================
XbarRouteUnitMflitRTL_test.py
==========================================================================
Unit tests for the multi-flit xbar route unit.

Author : Yanghui Ou
  Date : Feb 19, 2020
'''
import pytest
from pymtl3 import *
from pymtl3.stdlib.rtl.queues import BypassQueueRTL
from pymtl3.stdlib.test import mk_test_case_table
from ocnlib.utils import to_bits, to_bitstruct, run_sim
from ocnlib.test.test_srcs import MflitPacketSourceRTL as TestSource
from ocnlib.test.test_sinks import MflitPacketSinkRTL as TestSink
from ocnlib.packets import MflitPacket as Packet

from ..XbarRouteUnitMflitRTL import XbarRouteUnitMflitRTL

#-------------------------------------------------------------------------
# route_fl
#-------------------------------------------------------------------------

def route_fl( Header, num_outports, pkt_lst ):
  sink_pkts = [ [] for _ in range( num_outports ) ]
  for pkt in pkt_lst:
    header = to_bitstruct( pkt.flits[0], Header )
    dst    = header.dst.uint()
    sink_pkts[ dst ].append( pkt )
  return sink_pkts

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):

  def construct( s, Header, num_outports, pkts ):
    PhitType  = mk_bits( get_nbits( Header ) )
    sink_pkts = route_fl( Header, num_outports, pkts )

    s.src   = TestSource( Header, pkts )
    s.src_q = BypassQueueRTL( PhitType, num_entries=1 )
    s.dut   = XbarRouteUnitMflitRTL( Header, num_outports )
    s.sink  = [ TestSink( Header, sink_pkts[i] ) for i in range( num_outports ) ]

    s.src.send  //= s.src_q.enq
    s.src_q.deq //= s.dut.get

    for i in range( num_outports ):
      s.sink[i].recv.msg //= s.dut.give[i].ret
      s.sink[i].recv.en  //= lambda: s.dut.give[i].rdy & s.sink[i].recv.rdy
      s.dut.give[i].en   //= lambda: s.dut.give[i].rdy & s.sink[i].recv.rdy

  def done( s ):
    sinks_done = True
    for sink in s.sink:
      sinks_done &= sink.done()
    return s.src.done() and sinks_done

  def line_trace( s ):
    return s.dut.line_trace()

#-------------------------------------------------------------------------
# TestHeader
#-------------------------------------------------------------------------

@bitstruct
class TestHeader:
  src    : Bits8
  dst    : Bits8
  opaque : Bits8
  plen   : Bits8

#-------------------------------------------------------------------------
# mk_pkt
#-------------------------------------------------------------------------

def mk_pkt( src, dst, payload=[], opaque=0 ):
  plen   = len( payload )
  header = TestHeader( src, dst, opaque, plen )
  flits  = [ to_bits( header) ] + payload
  return Packet( TestHeader, flits )

#--------------------------------------------------------------------------
# test case: 1 pkt
#--------------------------------------------------------------------------

def one_pkt( num_outports ):
  return [ mk_pkt( 0, num_outports-1, [ 0x8badf00d, 0xfaceb00c ] ) ]

#--------------------------------------------------------------------------
# test case: 1 pkt to each
#--------------------------------------------------------------------------

def two_pkt_each( num_outports ):
  pkts = []

  for i in range( num_outports ):
    pkts.append( mk_pkt( 0, i, [ x for x in range( i ) ] ) )

  for i in range( num_outports ):
    pkts.append( mk_pkt( 0, i, [ x for x in range( (i+1) % 5 ) ] ) )

  return pkts

#--------------------------------------------------------------------------
# test case table
#--------------------------------------------------------------------------

test_cases = [
  (                      'msg_func      n_outs init flit_intv pkt_intv' ),
  [ '1pkt',               one_pkt,      4,     0,   0,        0         ],
  [ '1pkt_delay',         one_pkt,      4,     9,   3,        0         ],
  [ '2pkt_each_4',        two_pkt_each, 4,     0,   0,        9         ],
  [ '2pkt_each_4_delay',  two_pkt_each, 4,     4,   4,        0         ],
  [ '2pkt_each_6',        two_pkt_each, 6,     0,   0,        0         ],
  [ '2pkt_each_6_delay',  two_pkt_each, 6,     3,   6,        9         ],
  [ '2pkt_each_8',        two_pkt_each, 4,     0,   0,        0         ],
  [ '2pkt_each_8_delay',  two_pkt_each, 4,     8,   4,        9         ],
]

test_case_table = mk_test_case_table( test_cases )

#--------------------------------------------------------------------------
# test driver
#--------------------------------------------------------------------------

@pytest.mark.parametrize( **test_case_table )
def test_xbar_route( test_params, test_verilog ):
  pkts = test_params.msg_func( test_params.n_outs )
  th   = TestHarness( TestHeader, test_params.n_outs, pkts )
  th.set_param( 'top.sink*.construct',
    initial_delay         = test_params.init,
    flit_interval_delay   = test_params.flit_intv,
    packet_interval_delay = test_params.pkt_intv,
  )

  trans_backend = 'verilog' if test_verilog else ''
  run_sim( th, translation=trans_backend )



#=========================================================================
# ChannelCL_test.py
#=========================================================================
# Test for CL channel.
#
# Author : Yanghui Ou
#   Date : May 19, 2019

import pytest

from ocnlib.utils import run_sim
from pymtl3 import *
from pymtl3.stdlib.cl.queues import NormalQueueCL
from pymtl3.stdlib.test.test_sinks import TestSinkCL
from pymtl3.stdlib.test.test_srcs import TestSrcCL

from ..ChannelCL import ChannelCL

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):

  def construct( s, MsgType, src_msgs, sink_msgs ):

    s.src  = TestSrcCL ( MsgType, src_msgs  )
    s.sink = TestSinkCL( MsgType, sink_msgs )
    s.dut  = ChannelCL ( MsgType, latency=1 )

    # Connections
    s.src.send //= s.dut.recv
    s.dut.send //= s.sink.recv

  def done( s ):
    return s.src.done() and s.sink.done()

  def line_trace( s ):
    return s.src.line_trace() + "-> | " + s.dut.line_trace() + \
                               " | -> " + s.sink.line_trace()

#-------------------------------------------------------------------------
# Test cases
#-------------------------------------------------------------------------

test_msgs = [ b16(4), b16(1), b16(2), b16(3) ]

def test_chnl_simple():
  th = TestHarness( Bits16, test_msgs, test_msgs )
  run_sim( th )



'''
==========================================================================
MeshRouteUnitRTLMflitXY_test.py
==========================================================================
Unit tests for the multi-flit mesh route unit.

Author : Yanghui Ou
  Date : 11 Feb, 2020
'''
from pymtl3 import *
from pymtl3.stdlib.rtl.queues import BypassQueueRTL
from ocnlib.utils import to_bits, to_bitstruct, run_sim
from ocnlib.test.test_srcs import MflitPacketSourceRTL as TestSource
from ocnlib.test.test_sinks import MflitPacketSinkRTL as TestSink
from ocnlib.packets import MflitPacket as Packet

from ..directions import *
from ..MeshRouteUnitRTLMflitXY import MeshRouteUnitRTLMflitXY

#-------------------------------------------------------------------------
# Helper stuff
#-------------------------------------------------------------------------

def route_unit_fl( HeaderFormat, pkt_lst, pos_x, pos_y, first_dimension='x' ):
  sink_pkts = [ [] for _ in range(5) ]
  if first_dimension == 'x':
    for pkt in pkt_lst:
      header = to_bitstruct( pkt.flits[0], HeaderFormat )
      if header.dst_x == pos_x and header.dst_y == pos_y:
        sink_pkts[ SELF ].append( pkt )
      elif header.dst_x < pos_x:
        sink_pkts[ WEST ].append( pkt )
      elif header.dst_x > pos_x:
        sink_pkts[ EAST ].append( pkt )
      elif header.dst_y < pos_y:
        sink_pkts[ SOUTH ].append( pkt )
      else:
        sink_pkts[ NORTH ].append( pkt )

  elif first_dimension == 'y':
    for pkt in pkt_lst:
      header = to_bitstruct( pkt.flits[0], HeaderFormat )
      if header.dst_x == pos_x and header.dst_y == pos_y:
        sink_pkts[ SELF ].append( pkt )
      elif s.header.dst_y < pos_y:
        sink_pkts[ NORTH ].append( pkt )
      elif s.header.dst_y > pos_y:
        sink_pkts[ SOUTH ].append( pkt )
      elif s.header.dst_x < pos_x:
        sink_pkts[ WEST ].append( pkt )
      else:
        sink_pkts[ EAST ].append( pkt )

  else:
    assert False

  return sink_pkts

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):
  def construct( s, HeaderFormat, PositionType, pkts, x, y ):
    PhitType = mk_bits( get_nbits( HeaderFormat ) )
    sink_pkts = route_unit_fl( HeaderFormat, pkts, x, y )

    s.src   = TestSource( HeaderFormat, pkts )
    s.src_q = BypassQueueRTL( PhitType, num_entries=1 )
    s.dut   = MeshRouteUnitRTLMflitXY( HeaderFormat, PositionType )
    s.sink  = [ TestSink( HeaderFormat, sink_pkts[i] ) for i in range(5) ]

    s.src.send  //= s.src_q.enq
    s.src_q.deq //= s.dut.get
    s.dut.pos   //= PositionType( x, y )

    for i in range(5):
      s.sink[i].recv.msg //= s.dut.give[i].ret
      s.sink[i].recv.en  //= lambda: s.dut.give[i].rdy & s.sink[i].recv.rdy
      s.dut.give[i].en   //= lambda: s.dut.give[i].rdy & s.sink[i].recv.rdy

  def done( s ):
    sinks_done = True
    for sink in s.sink:
      sinks_done &= sink.done()
    return s.src.done() and sinks_done

  def line_trace( s ):
    return s.dut.line_trace()

#-------------------------------------------------------------------------
# TestHeader
#-------------------------------------------------------------------------

@bitstruct
class TestHeader:
  opaque: Bits8
  plen  : Bits8
  src_x : Bits4
  src_y : Bits4
  dst_x : Bits4
  dst_y : Bits4

#-------------------------------------------------------------------------
# TestPosition
#-------------------------------------------------------------------------

@bitstruct
class TestPosition:
  pos_x : Bits4
  pos_y : Bits4

#-------------------------------------------------------------------------
# mk_pkt
#-------------------------------------------------------------------------

def mk_pkt( dst_x, dst_y, payload=[], src_x=0, src_y=0, opaque=0 ):
  plen        = len( payload )
  header      = TestHeader( opaque, plen, src_x, src_y, dst_x, dst_y )
  header_bits = to_bits( header )
  flits       = [ header_bits ] + payload
  return Packet( TestHeader, flits )

#-------------------------------------------------------------------------
# test_1pkt
#-------------------------------------------------------------------------

def test_1pkt():
  pkts= [
    mk_pkt( 0, 1, [ 0xfaceb00c, 0x8badf00d ] ),
  ]
  th = TestHarness( TestHeader, TestPosition, pkts, 0, 0 )
  run_sim( th, max_cycles=20 )

#-------------------------------------------------------------------------
# test_4pkt
#-------------------------------------------------------------------------

def test_4pkt():
  pkts= [
    mk_pkt( 1, 0, [                                    ] ),
    mk_pkt( 1, 2, [ 0xfaceb00c, 0x8badf00d             ] ),
    mk_pkt( 0, 1, [ 0xdeadbeef                         ] ),
    mk_pkt( 2, 1, [ 0xcafef00d, 0x111edcba, 0xbaaaaaad ] ),
  ]
  th = TestHarness( TestHeader, TestPosition, pkts, 1, 1 )
  th.set_param( 'top.src.construct', packet_interval_delay = 1 )
  run_sim( th, max_cycles=20 )



'''
==========================================================================
SwitchUnitGrantHoldRTL_test.py
==========================================================================

Author : Yanghui Ou
  Date : Jan 28, 2020
'''
from pymtl3 import *
from ..SwitchUnitGrantHoldRTL import SwitchUnitGrantHoldRTL

def test_sanity_check():
  dut = SwitchUnitGrantHoldRTL( Bits16, num_inports=5 )
  dut.elaborate()
  dut.apply( SimulationPass() )
  dut.sim_reset()
  dut.tick()

#-------------------------------------------------------------------------
# test_adhoc
#-------------------------------------------------------------------------
# A simple adhoc test. Port0 sends 2 packets, [ face, b000c, 8bad, f00d ]
# and [ c001 ]; while port1 sends one packet [ dead, beef ].

def test_adhoc():
  dut = SwitchUnitGrantHoldRTL( Bits16, num_inports=5 )
  dut.elaborate()
  dut.apply( SimulationPass() )
  dut.sim_reset()
  print()

  # Cycle 1 - first flit, no competitor
  dut.get[0].rdy = b1(1)
  dut.get[0].ret = b16( 0xface )
  dut.hold[0]    = b1(0)
  dut.give.en    = b1(1)
  dut.eval_combinational()
  assert dut.give.rdy
  assert dut.give.ret == 0xface
  print( dut.line_trace() )
  dut.tick()

  # Cycle 2 - second flit, set hold logic, another flit comes in
  dut.get[0].rdy = b1(1)
  dut.get[0].ret = b16( 0xb00c )
  dut.hold[0]    = b1(1)

  dut.get[1].rdy = b1(1)
  dut.get[1].ret = b16( 0x0001 )
  dut.hold[1]    = b1(0)

  dut.give.en    = b1(1)
  dut.eval_combinational()
  print( dut.line_trace() )
  assert dut.give.rdy
  assert dut.give.ret == 0xb00c
  dut.tick()

  # Cycle 3 - third flit from port0
  dut.get[0].rdy = b1(1)
  dut.get[0].ret = b16( 0x8bad )
  dut.hold[0]    = b1(1)

  dut.get[1].rdy = b1(1)
  dut.get[1].ret = b16( 0xdead )
  dut.hold[1]    = b1(0)

  dut.give.en    = b1(1)
  dut.eval_combinational()
  print( dut.line_trace() )
  assert dut.give.rdy
  assert dut.give.ret == 0x8bad
  dut.tick()

  # Cycle 4 - last flit from port0
  dut.get[0].rdy = b1(1)
  dut.get[0].ret = b16( 0xf00d )
  dut.hold[0]    = b1(1)

  dut.get[1].rdy = b1(1)
  dut.get[1].ret = b16( 0xdead )
  dut.hold[1]    = b1(0)

  dut.give.en    = b1(1)
  dut.eval_combinational()
  print( dut.line_trace() )
  assert dut.give.rdy
  assert dut.give.ret == 0xf00d
  dut.tick()

  # Cycle 5 - arbitrate header
  dut.get[0].rdy = b1(1)
  dut.get[0].ret = b16( 0xc001 )
  dut.hold[0]    = b1(0)

  dut.get[1].rdy = b1(1)
  dut.get[1].ret = b16( 0xdead )
  dut.hold[1]    = b1(0)

  dut.give.en    = b1(1)
  dut.eval_combinational()
  print( dut.line_trace() )
  assert dut.give.rdy
  assert dut.give.ret == 0xdead
  dut.tick()

  # Cycle 5 - arbitrate header
  dut.get[0].rdy = b1(1)
  dut.get[0].ret = b16( 0xc001 )
  dut.hold[0]    = b1(0)

  dut.get[1].rdy = b1(1)
  dut.get[1].ret = b16( 0xbeef )
  dut.hold[1]    = b1(1)

  dut.give.en    = b1(1)
  dut.eval_combinational()
  print( dut.line_trace() )
  assert dut.give.rdy
  assert dut.give.ret == 0xbeef
  dut.tick()

  # Cycle 6 - single flit packet from port0
  dut.get[0].rdy = b1(1)
  dut.get[0].ret = b16( 0xc001 )
  dut.hold[0]    = b1(0)

  dut.get[1].rdy = b1(0)
  dut.get[1].ret = b16( 0xbeef )
  dut.hold[1]    = b1(0)

  dut.give.en    = b1(1)
  dut.eval_combinational()
  print( dut.line_trace() )
  assert dut.give.rdy
  assert dut.give.ret == 0xc001
  dut.tick()




"""
==========================================================================
InputUnitCL.py
==========================================================================
Cycle level implementeation of the CL model.

Author : Yanghui Ou
  Date : May 16, 2019
"""
from pymtl3 import *
from pymtl3.stdlib.cl.queues import NormalQueueCL


class InputUnitCL( Component ):

  def construct( s, PacketType, QueueType=NormalQueueCL ):

    # Interface

    s.recv = CalleeIfcCL( Type=PacketType )
    s.give = CalleeIfcCL( Type=PacketType )

    # Component

    s.queue = QueueType( num_entries=2 )( enq = s.recv, deq = s.give )

  def line_trace( s ):
    return f"{s.recv}(){s.give}"



#=========================================================================
# MeshRouteRTL.py
#=========================================================================
# Simple network-on-chip router, try to connect all the units together
#
# Author : Cheng Tan, Yanghui Ou
#   Date : Mar 8, 2019

from cmeshnet.DORYCMeshRouteUnitRTL import DORYCMeshRouteUnitRTL
from pymtl3 import *
from router.InputUnitRTL import InputUnitRTL
from router.OutputUnitRTL import OutputUnitRTL
from router.Router import Router
from router.SwitchUnitRTL import SwitchUnitRTL


class CMeshRouterRTL( Router ):

  def construct( s, PacketType, PositionType, num_inports, num_outports,
                 InputUnitType  = InputUnitRTL,
                 RouteUnitType  = DORYCMeshRouteUnitRTL,
                 SwitchUnitType = SwitchUnitRTL ):

    super().construct(
      PacketType,
      PositionType,
      num_inports,
      num_outports,
      InputUnitType,
      RouteUnitType,
      SwitchUnitType,
      OutputUnitRTL,
    )



'''
==========================================================================
test_srcs.py
==========================================================================
A collection of test sources.

Author : Yanghui Ou
  Date : Feb 2, 2020
'''
from collections import deque
from pymtl3 import *
from pymtl3.stdlib.ifcs import RecvCL2SendRTL, SendIfcRTL

from ..utils import get_nbits

#-------------------------------------------------------------------------
# MflitPacketSourceCL
#-------------------------------------------------------------------------
# pkts                : a list of MflitPacket objects.
# initial_delay       : number of cycles before sending the very first flit.
# flit_interval_delay : number of cycles between each flit in a packet.
# packet_interval_delay  : number of cycles between each packet.
# TODO: check if inputs packtes are valid

class MflitPacketSourceCL( Component ):

  def construct( s, Format, pkts, initial_delay=0, flit_interval_delay=0, packet_interval_delay=0 ):

    # Interface
    PhitType = mk_bits( get_nbits( Format ) )
    s.send = CallerIfcCL( Type=PhitType )

    # Metadata
    s.pkts    = deque( pkts )
    s.cur_pkt = None
    s.count   = initial_delay
    s.f_delay = flit_interval_delay
    s.p_delay = packet_interval_delay

    # Update block
    @s.update
    def up_src_send():
      if s.count > 0:
        s.count -= 1
      elif not s.reset:
        # pop a packet to send
        if not s.cur_pkt and s.pkts:
          s.cur_pkt = s.pkts.popleft()
          assert not s.cur_pkt.empty()

        if s.send.rdy() and s.cur_pkt:
          s.send( s.cur_pkt.pop() )

          if s.cur_pkt.empty():
            s.cur_pkt = None
            s.count   = s.p_delay
          else:
            s.count   = s.f_delay

  def done( s ):
    return not s.pkts and not s.cur_pkt

  def line_trace( s ):
    return f'({s.count}){s.send}'

#-------------------------------------------------------------------------
# MflitPacketSourceRTL
#-------------------------------------------------------------------------

class MflitPacketSourceRTL( Component ):

  def construct( s, Format, pkts, initial_delay=0, flit_interval_delay=0,
                 packet_interval_delay=0, cmp_fn=lambda a, b : a.flits == b.flits ):

    PhitType = mk_bits( get_nbits( Format ) )

    s.send    = SendIfcRTL( PhitType )
    s.src_cl  = MflitPacketSourceCL( Format, pkts, initial_delay, flit_interval_delay,
                                         packet_interval_delay )
    s.adapter = RecvCL2SendRTL( PhitType )

    connect( s.src_cl.send,  s.adapter.recv )
    connect( s.adapter.send, s.send         )

  def done( s ):
    return s.src_cl.done()

  def line_trace( s ):
    return f'{s.send}'



"""
==========================================================================
RouteUnitDorFL.py
==========================================================================
FL route unit that implements dimension order routing.

Author : Yanghui Ou
  Date : June 30, 2019
"""
from pymtl3 import *

from .directions import *


class RouteUnitDorFL:

  def __init__( s, pos_x, pos_y, ncols, nrows, dimension='y' ):
    s.pos_x = pos_x
    s.pos_y = pos_y
    s.ncols = ncols
    s.nrows = nrows
    s.dimension   = dimension
    s.last_row_id = nrows-1
    s.last_col_id = ncols-1

  def route( s, src_pkts ):
    dst_pkts = [ [] for _ in range(5) ]

    for pkt in src_pkts:
      if pkt.dst_x == s.pos_x and pkt.dst_y == s.pos_y:
        dst_pkts[ SELF ].append( pkt )
      else:
        north_dist = pkt.dst_y - s.pos_y if pkt.dst_y > s.pos_y else pkt.dst_y + s.nrows - s.pos_y
        south_dist = s.pos_y - pkt.dst_y if pkt.dst_y < s.pos_y else s.pos_y + s.nrows - pkt.dst_y
        east_dist  = pkt.dst_x - s.pos_x if pkt.dst_x > s.pos_x else pkt.dst_x + s.ncols - s.pos_x
        west_dist  = s.pos_x - pkt.dst_x if pkt.dst_x < s.pos_x else s.pos_x + s.ncols - pkt.dst_x

        if s.dimension == 'y':
          if pkt.dst_y != s.pos_y:
            out_dir = NORTH if north_dist < south_dist else SOUTH
          else:
            out_dir = WEST if west_dist < east_dist else EAST
          dst_pkts[ out_dir ].append( pkt )

        else: # s.dimension=='x'
          if pkt.dst_x != s.pos_x:
            out_dir = WEST if west_dist < east_dist else EAST
          else:
            out_dir = NORTH if north_dist < south_dist else SOUTH
          dst_pkts[ out_dir ].append( pkt )

    return dst_pkts



"""
==========================================================================
OutputUnitCreditRTL.py
==========================================================================
An output unit with a credit based interface.

Author : Yanghui Ou
  Date : June 22, 2019
"""
from ocnlib.ifcs.CreditIfc import CreditSendIfcRTL
from ocnlib.rtl import Counter
from pymtl3 import *
from pymtl3.stdlib.ifcs import GetIfcRTL, enrdy_to_str


class OutputUnitCreditRTL( Component ):

  def construct( s, MsgType, vc=2, credit_line=2 ):
    assert vc > 1

    # Interface
    s.get = GetIfcRTL( MsgType )
    s.send = CreditSendIfcRTL( MsgType, vc )

    s.MsgType = MsgType
    s.vc    = vc

    # Loval types
    credit_type = mk_bits( clog2(credit_line+1) )
    vcid_type   = mk_bits( clog2( vc ) if vc > 1 else 1 )

    s.credit = [ Counter( credit_type, credit_line ) for _ in range( vc ) ]

    s.get.ret //= s.send.msg

    @s.update
    def up_credit_send():
      s.send.en = b1(0)
      s.get.en = b1(0)
      if s.get.rdy:
        # print( str(s) + " : " + str(s.get.ret) )
        for i in range( vc ):
          if vcid_type(i) == s.get.ret.vc_id and s.credit[i].count > credit_type(0):
            s.send.en = b1(1)
            s.get.en = b1(1)

    @s.update
    def up_counter_decr():
      for i in range( vc ):
        s.credit[i].decr = s.send.en & ( vcid_type(i) == s.send.msg.vc_id )

    for i in range( vc ):
      s.credit[i].incr       //= s.send.yum[i]
      s.credit[i].load       //= b1(0)
      s.credit[i].load_value //= credit_type(0)

  def line_trace( s ):
    return "{}({}){}".format(
      s.get,
      ",".join( [ str(s.credit[i].count) for i in range(s.vc) ] ),
      s.send,
    )



'''
==========================================================================
BoundaryUnit_test.py
==========================================================================
Unit test for the boundary unit.

Author : Yanghui Ou
  Date : Feb 17, 2020
'''
from pymtl3 import *
from .BoundaryUnit import BoundaryUnit

def test_false():
  m = BoundaryUnit( default_rdy=False )
  m.elaborate()
  m.apply( SimulationPass() )
  m.tick()
  assert not m.recv.rdy()

  try:
    m.recv( 1 )
    flag = False
  except AssertionError:
    flag = True
  
  assert flag

def test_true():
  m = BoundaryUnit( default_rdy=True )
  m.elaborate()
  m.apply( SimulationPass() )
  m.tick()
  assert m.recv.rdy()

  try:
    m.recv( 1 )
    flag = False
  except AssertionError:
    flag = True
  
  assert flag



'''
==========================================================================
SerializerRTL.py
==========================================================================
A generic serializer unit.

Author : Yanghui Ou
  Date : Feb 25, 2020
'''
from pymtl3 import *
from pymtl3.stdlib.ifcs import SendIfcRTL, RecvIfcRTL
from pymtl3.stdlib.rtl import Mux

from .Counter import Counter

class SerializerRTL( Component ):

  #-----------------------------------------------------------------------
  # Construct
  #-----------------------------------------------------------------------

  def construct( s, out_nbits, max_nblocks ):

    # Local parameter

    InType    = mk_bits( out_nbits*( max_nblocks ) )
    OutType   = mk_bits( out_nbits )
    CountType = mk_bits( clog2( max_nblocks+1 ) )

    s.STATE_IDLE = b1(0)
    s.STATE_SEND = b1(1)

    s.sel_nbits = clog2( max_nblocks )

    # Interface

    s.recv = RecvIfcRTL( InType    )
    s.send = SendIfcRTL( OutType   )
    s.len  = InPort    ( CountType )

    # Components

    s.state      = Wire( Bits1     )
    s.state_next = Wire( Bits1     )
    s.in_r       = Wire( InType    )
    s.len_r      = Wire( CountType )

    s.counter    = Counter( CountType )( decr=b1(0) )
    s.mux        = Mux( OutType, max_nblocks )

    # Input register

    @s.update_ff
    def up_in_r():
      if s.recv.en & ( s.state_next != s.STATE_IDLE ):
        s.in_r  <<= s.recv.msg
        # Force len to 1 if it is set to be 0 to avoid undefined behavior
        s.len_r <<= s.len if s.len > CountType(0) else CountType(1)
      else:
        s.in_r  <<= s.in_r
        s.len_r <<= CountType(0) if s.state_next == s.STATE_IDLE else s.len_r

    # Mux logic

    for i in range( max_nblocks ):
      s.mux.in_[i] //= s.in_r[ i*out_nbits : (i+1)*out_nbits ]
    s.mux.sel //= s.counter.count[0:s.sel_nbits]

    # Counter load

    s.counter.load //= lambda: s.recv.en | ( s.state_next == s.STATE_IDLE )
    s.counter.incr //= lambda: s.send.en & ( s.state_next != s.STATE_IDLE )

    @s.update
    def up_counter_load_value():
      if ( s.state == s.STATE_IDLE ) & s.send.rdy & ( s.state_next != s.STATE_IDLE ):
        s.counter.load_value = CountType(1)
      else:
        s.counter.load_value = CountType(0)

    # Recv logic

    s.recv.rdy //= lambda: s.state == s.STATE_IDLE
    
    # Send logic

    @s.update
    def up_send_msg():
      if ( s.state == s.STATE_IDLE ) & s.recv.en & s.send.rdy:
        s.send.msg = s.recv.msg[0:out_nbits]
      else:
        s.send.msg = s.mux.out

    @s.update
    def up_send_en():
      if ( s.state == s.STATE_IDLE ) & s.recv.en & s.send.rdy | \
         ( s.state == s.STATE_SEND ) & s.send.rdy:
        s.send.en = b1(1)
      else:
        s.send.en = b1(0)

    # State transition logic

    @s.update_ff
    def up_state():
      if s.reset:
        s.state <<= s.STATE_IDLE
      else:
        s.state <<= s.state_next

    @s.update
    def up_state_next():
      if s.state == s.STATE_IDLE:
        # If length is 1, bypass to IDLE
        if ( s.len == CountType(1) ) & s.send.en:
          s.state_next = s.STATE_IDLE

        elif s.recv.en:
          s.state_next = s.STATE_SEND

        else:
          s.state_next = s.STATE_IDLE

      else: # STATE_SEND
        if ( s.counter.count == s.len_r - CountType(1) ) & s.send.rdy:
          s.state_next = s.STATE_IDLE
        else:
          s.state_next = s.STATE_SEND

  #-----------------------------------------------------------------------
  # line trace
  #-----------------------------------------------------------------------

  def line_trace( s ):
    state = 'I' if s.state == s.STATE_IDLE else \
            'S' if s.state == s.STATE_SEND else \
            '?'
    state_next = 'I' if s.state_next == s.STATE_IDLE else \
            'S' if s.state_next == s.STATE_SEND else \
            '?'
    return f'{s.recv}({state}{s.counter.count}<{s.len_r}){s.send}'



"""
=========================================================================
OutputUnitCL_test.py
=========================================================================
Test cases for OutputUnitCL.

Author : Yanghui Ou
  Date : Feb 28, 2019
"""
import hypothesis
import pytest
from hypothesis import strategies as st

from pymtl3 import *
from pymtl3.datatypes import strategies as pst
from pymtl3.stdlib.cl.queues import BypassQueueCL, NormalQueueCL, PipeQueueCL
from pymtl3.stdlib.test.test_sinks import TestSinkCL
from pymtl3.stdlib.test.test_srcs import TestSrcCL
from router.OutputUnitCL import OutputUnitCL

from .InputUnitCL_test import InputUnitCL_Tests as BaseTests

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):

  def construct( s, MsgType, src_msgs, sink_msgs ):

    s.src   = TestSrcCL( MsgType, src_msgs  )
    s.src_q = BypassQueueCL( num_entries=1 )
    s.dut   = OutputUnitCL( MsgType )
    s.sink  = TestSinkCL( MsgType, sink_msgs )

    # Connections
    s.src.send  //= s.src_q.enq
    s.src_q.deq //= s.dut.get
    s.dut.send  //= s.sink.recv

  def done( s ):
    return s.src.done() and s.sink.done()

  def line_trace( s ):
    return "{} {} {}".format(
      s.src.line_trace(),
      s.dut.line_trace(),
      s.sink.line_trace(),
    )

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------


class OutputUnitCL_Tests( BaseTests ):

  @classmethod
  def setup_class( cls ):
    cls.TestHarness = TestHarness
    cls.qtypes      = [ NormalQueueCL, BypassQueueCL, PipeQueueCL ]



"""
==========================================================================
MeshRouterCL_test.py
==========================================================================
Test cases for MeshRouterCL.

Author : Yanghui Ou
  Date : May 21, 2019
"""
from itertools import product

import hypothesis
import pytest
from hypothesis import strategies as st

from meshnet.DORXMeshRouteUnitCL import DORXMeshRouteUnitCL
from meshnet.MeshRouterCL import MeshRouterCL
from meshnet.MeshRouterFL import MeshRouterFL
from ocnlib.ifcs.packets import mk_mesh_pkt
from ocnlib.ifcs.positions import mk_mesh_pos
from ocnlib.utils import run_sim
from ocnlib.test.net_sinks import TestNetSinkCL as TestSinkCL
from pymtl3 import *
from pymtl3.datatypes import strategies as pst
from pymtl3.stdlib.test.test_srcs import TestSrcCL

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):

  def construct( s,
    MsgType   = None,
    src_msgs  = [],
    sink_msgs = [],
    ncols  = 2,
    nrows   = 2 ,
    pos_x     = 0,
    pos_y     = 0,
  ):

    MeshPos = mk_mesh_pos( ncols, nrows )
    s.nrouters = ncols * nrows
    match_func = lambda a, b : a.src_x == a.src_x and a.src_y == b.src_y and \
                               a.dst_x == b.dst_x and a.dst_y == b.dst_y and \
                               a.opaque == b.opaque and a.payload == b.payload
    s.dut = MeshRouterCL( MsgType, MeshPos )

    s.srcs  = [ TestSrcCL( MsgType, src_msgs[i] )
                for i in range(5) ]
    s.sinks = [ TestSinkCL( MsgType, sink_msgs[i], match_func=match_func )
                for i in range(5) ]

    # Connections

    for i in range ( s.dut.num_outports ):
      s.srcs[i].send //= s.dut.recv[i]
      s.dut.send[i]  //= s.sinks[i].recv

    @s.update
    def up_pos():
      s.dut.pos = MeshPos( pos_x, pos_y )

  def done( s ):
    srcs_done = 1
    sinks_done = 1
    for i in range( s.dut.num_inports ):
      if s.srcs[i].done() == 0:
        srcs_done = 0
    for i in range( s.dut.num_outports ):
      if s.sinks[i].done() == 0:
        sinks_done = 0
    return srcs_done and sinks_done

  def line_trace( s ):
    return "{}".format( s.dut.line_trace() )

#-------------------------------------------------------------------------
# Test cases
#-------------------------------------------------------------------------

@st.composite
def mesh_pkt_strat( draw, ncols, nrows, opaque_nbits=8, vc=1, payload_nbits=32 ):
  dst_x = draw( st.integers(0, ncols-1) )
  dst_y = draw( st.integers(0, nrows -1) )
  src_x = draw( st.integers(0, ncols-1) )
  src_y = draw( st.integers(0, nrows -1) )
  opaque  = draw( pst.bits( opaque_nbits ) )
  payload = draw( st.sampled_from([ 0, 0xdeadbeef, 0xfaceb00c, 0xc001cafe ]) )
  Pkt = mk_mesh_pkt( ncols, nrows, opaque_nbits, vc, payload_nbits )
  if vc==1:
    return Pkt( src_x, src_y, dst_x, dst_y, opaque, payload )
  else:
    return Pkt( src_x, src_y, dst_x, dst_y, opaque, 0, payload )

#-------------------------------------------------------------------------
# Test cases
#-------------------------------------------------------------------------

class MeshRouterCL_Tests:

  @classmethod
  def setup_class( s ):
    s.TestHarness = TestHarness

  @pytest.mark.parametrize(
    'pos_x, pos_y',
    product( [ 0, 1, 2, 3 ], [ 0, 1, 2, 3 ] ),
  )
  def test_simple4x4( s, pos_x, pos_y ):
    ncols = 4
    nrows  = 4
    Pkt = mk_mesh_pkt( ncols, nrows )
    router_fl = MeshRouterFL( pos_x, pos_y, dimension='x' )
    src_pkts = router_fl.arrange_src_pkts([
      #   src_x  y dst_x  y  opq payload
      Pkt(    0, 0,    3, 3, 0,  0xdeaddead ),
    ])
    dst_pkts = router_fl.route( src_pkts )
    th = s.TestHarness( Pkt, src_pkts, dst_pkts, ncols, nrows, pos_x, pos_y )
    run_sim( th )

  # Failing test cases captured by hypothesis

  def test_h0( s ):
    pos_x = 0; pos_y = 0; ncols = 2; nrows = 2
    Pkt = mk_mesh_pkt( ncols, nrows )
    router_fl = MeshRouterFL( pos_x, pos_y, dimension='x' )
    src_pkts = router_fl.arrange_src_pkts([
      #   src_x  y dst_x  y  opq payload
      Pkt(    0, 0,    1, 0, 0,  0xdeadbabe ),
      Pkt(    0, 0,    1, 0, 0,  0xdeadface ),
    ])
    dst_pkts = router_fl.route( src_pkts )
    th = s.TestHarness( Pkt, src_pkts, dst_pkts, ncols, nrows, pos_x, pos_y )
    run_sim( th )

  def test_h1( s ):
    pos_x = 0; pos_y = 0; ncols = 2; nrows = 2
    Pkt = mk_mesh_pkt( ncols, nrows )
    router_fl = MeshRouterFL( pos_x, pos_y, dimension='x' )
    src_pkts = router_fl.arrange_src_pkts([
      #   src_x  y dst_x  y  opq payload
      Pkt(    0, 0,    0, 1, 0,  0xdeadbabe ),
    ])
    dst_pkts = router_fl.route( src_pkts )
    th = s.TestHarness( Pkt, src_pkts, dst_pkts, ncols, nrows, pos_x, pos_y )
    run_sim( th )

#-------------------------------------------------------------------------
# Hypothesis test
#-------------------------------------------------------------------------

  @hypothesis.settings( deadline = None )
  @hypothesis.given(
    ncols   = st.integers(2, 16),
    nrows    = st.integers(2, 16),
    routing    = st.sampled_from(['x']), # TODO: add y after implementing DORY route unit
    pos_x      = st.data(),
    pos_y      = st.data(),
    pkts       = st.data(),
    src_init   = st.integers(0, 2 ),
    src_inter  = st.integers(0, 5 ),
    sink_init  = st.integers(0, 20),
    sink_inter = st.integers(0, 5 ),
  )
  def test_hypothesis( s, ncols, nrows, routing, pos_x, pos_y, pkts,
      src_init, src_inter, sink_init, sink_inter ):
    # Draw some numbers
    pos_x = pos_x.draw( st.integers(0,ncols-1), label="pos_x" )
    pos_y = pos_y.draw( st.integers(0,nrows-1), label="pos_y" )
    Pkt   = mk_mesh_pkt( ncols, nrows )
    router_fl = MeshRouterFL( pos_x, pos_y, dimension='x' )
    msgs  = pkts.draw(
      st.lists( mesh_pkt_strat( ncols, nrows ), min_size = 1, max_size = 50 ),
      label = "msgs"
    )
    src_pkts = router_fl.arrange_src_pkts( msgs )
    dst_pkts = router_fl.route( src_pkts )
    th = s.TestHarness( Pkt, src_pkts, dst_pkts, ncols, nrows, pos_x, pos_y )
    th.set_param( "top.src*.construct", initial_delay=src_init, interval_delay=src_init )
    th.set_param( "top.sink*.construct", initial_delay=sink_init, interval_delay=sink_init )
    run_sim( th, max_cycles=5000 )



#=========================================================================
# Router.py
#=========================================================================
# A generic router model.
#
# Author : Yanghui Ou, Cheng Tan
#   Date : Mar 26, 2019

from ocnlib.ifcs.PhysicalDimension import PhysicalDimension
from pymtl3 import *
from pymtl3.stdlib.ifcs import RecvIfcRTL, SendIfcRTL


class Router( Component ):

  def construct( s, PacketType, PositionType, num_inports, num_outports,
                 InputUnitType, RouteUnitType, SwitchUnitType,
                 OutputUnitType ):

    s.dim = PhysicalDimension()
    s.num_inports  = num_inports
    s.num_outports = num_outports

    # Interface

    s.pos  = InPort( PositionType )
    s.recv = [ RecvIfcRTL( PacketType ) for _ in range( s.num_inports  ) ]
    s.send = [ SendIfcRTL( PacketType ) for _ in range( s.num_outports ) ]

    # Components

    s.input_units  = [ InputUnitType( PacketType )
                      for _ in range( s.num_inports ) ]

    s.route_units  = [ RouteUnitType( PacketType, PositionType, s.num_outports )
                      for i in range( s.num_inports ) ]

    s.switch_units = [ SwitchUnitType( PacketType, s.num_inports )
                      for _ in range( s.num_outports ) ]

    s.output_units = [ OutputUnitType( PacketType )
                      for _ in range( s.num_outports ) ]

    # Connection

    for i in range( s.num_inports ):
      s.recv[i]             //= s.input_units[i].recv
      s.input_units[i].give //= s.route_units[i].get
      s.pos                 //= s.route_units[i].pos

    for i in range( s.num_inports ):
      for j in range( s.num_outports ):
        s.route_units[i].give[j] //= s.switch_units[j].get[i]

    for j in range( s.num_outports ):
      s.switch_units[j].give //= s.output_units[j].get
      s.output_units[j].send //= s.send[j]

  # Line trace

  def line_trace( s ):
    return "{}({}){}".format(
      "|".join( [ f"{x}" for x in s.recv ] ),
      s.pos,
      "|".join( [ f"{x}" for x in s.send ] )
    )

  def elaborate_physical( s ):
    s.dim.w = 50
    s.dim.h = 150
